﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Globalization;

namespace CodeIT_S6_Visualizer
{
    public partial class Visualizer : Form
    {
        public static void Debug(string message = "Test")
        {
            MessageBox.Show(message, "Debug", MessageBoxButtons.OK, MessageBoxIcon.Information);
        } // Wrapper for MessageBox.Show(). Should be used for debug purposes.
        public static void Debug_WriteFile<T>(T container, string filename) where T : IEnumerable<T>
        {
            // Set a variable to the My Documents path.
                    string mydocpath = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);

                    // Write the string array to a new file named "WriteLines.txt".
                    using (StreamWriter outputFile = new StreamWriter(mydocpath + filename))
                    {
                        foreach (T line in container)
                            outputFile.WriteLine(line);
                    }
        } // Writes contents of T to file filename in My Documents
        public static void ReportError(string error_message)
        {
            MessageBox.Show(error_message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
        } // Wrapper for MessageBox.Show(). Should be used when reporting errors.

        static InputProcessor InputFile = new InputProcessor(); // Processor for input file
        static OutputProcessor OutputFile = new OutputProcessor(); // Processor for output file
        static GridInfo Grid = new GridInfo(); // Information on the current state of the grid
        static Stack<Vector> LatestChanges = new Stack<Vector>(); // Contains the latest updated points
        static bool FinishedVisualization = false;

        public Visualizer()
        {
            InitializeComponent();
        } // Startup function

#pragma warning disable CS0661 // Type defines operator == or operator != but does not override Object.GetHashCode()
#pragma warning disable CS0660 // Type defines operator == or operator != but does not override Object.Equals(object o)
        public struct Vector 
#pragma warning restore CS0660 // Type defines operator == or operator != but does not override Object.Equals(object o)
#pragma warning restore CS0661 // Type defines operator == or operator != but does not override Object.GetHashCode()
        {
            public int x, y;

            public Vector(int _x, int _y) // Constructor
            {
                x = _x;
                y = _y;
            }

            public Point ToPoint()
            {
                return new Point(x, y);
            }

            public static Vector operator +(Vector A, Vector B) // Add vectors
            {
                return new Vector(A.x + B.x, A.y + B.y);
            }

            public static bool operator == (Vector A, Vector B) // Comparer
            {
                return A.x == B.x && A.y == B.y;
            }

            public static bool operator !=(Vector A, Vector B) // Comparer
            {
                return A.x != B.x || A.y != B.y;
            }

            public static Vector operator -(Vector A, Vector B) // Substract vectors
            {
                return new Vector(A.x - B.x, A.y - B.y);
            }

            public static Vector operator *(Vector A, int lambda) // Multiply a vector by a scalar
            {
                return new Vector(A.x * lambda, A.y * lambda);
            }

            public static int operator ^(Vector A, Vector B) // cross/vector product
            {
                return A.x * B.y - B.x * A.y;
            }

            public Vector Rotate(int a, int b, int c, int d) // rotates a vector according to the matrix with elements a, b, c, d
            {
                return new Vector(a * this.x + b * this.y, c * this.x + d * this.y);
            }

            public double Length() // the length of the vector
            {
                return Math.Sqrt(this.x * this.x + this.y * this.y);
            }
        } // This class is used for representing points and vectors

#pragma warning disable CS0660 // Type defines operator == or operator != but does not override Object.Equals(object o)
#pragma warning disable CS0661 // Type defines operator == or operator != but does not override Object.GetHashCode()
        public struct DoubleVector
#pragma warning restore CS0661 // Type defines operator == or operator != but does not override Object.GetHashCode()
#pragma warning restore CS0660 // Type defines operator == or operator != but does not override Object.Equals(object o)
        {
            public double x, y;

            public DoubleVector(double _x, double _y)
            {
                x = _x;
                y = _y;
            }

            public static DoubleVector operator +(DoubleVector A, DoubleVector B) // Add vectors
            {
                return new DoubleVector(A.x + B.x, A.y + B.y);
            }

            public static bool operator ==(DoubleVector A, DoubleVector B) // Comparer
            {
                return A.x == B.x && A.y == B.y;
            }

            public static bool operator !=(DoubleVector A, DoubleVector B) // Comparer
            {
                return A.x != B.x || A.y != B.y;
            }

            public static DoubleVector operator -(DoubleVector A, DoubleVector B) // Substract vectors
            {
                return new DoubleVector(A.x - B.x, A.y - B.y);
            }

            public static DoubleVector operator *(DoubleVector A, double lambda) // Multiply a vector by a scalar
            {
                return new DoubleVector(A.x * lambda, A.y * lambda);
            }

            public static double operator ^(DoubleVector A, DoubleVector B) // cross/vector product
            {
                return A.x * B.y - B.x * A.y;
            }

            public DoubleVector Rotate(double a, double b, double c, double d) // rotates a vector according to the matrix with elements a, b, c, d
            {
                return new DoubleVector(a * this.x + b * this.y, c * this.x + d * this.y);
            }

            public double Length() // the length of the vector
            {
                return Math.Sqrt(this.x * this.x + this.y * this.y);
            }
        } // Same as Vector but working with double (blame C#)

        public class Lines 
        {
            public Vector A, B, C, D; // A, B define the lower line; C, D define the upper line
            public double dist;
        } // This class is used to store the two lines restricting the points and the distance between them
        public struct VectorEvent
        {
            public enum Operation { DEFAULT = 0, HOMOTETHY = 1, TWO = 2, ADD = 3 };
            public enum Effect { REMOVED = -1, NEUTRAL = 0, ADDED = 1 };
            public Color GetColor()
            {
                switch (operation)
                {
                    case Operation.HOMOTETHY:
                        switch (effect)
                        {
                            case Effect.REMOVED:
                                return Color.Red;
                            case Effect.NEUTRAL:
                                return Color.Blue;
                            case Effect.ADDED:
                                return Color.Green;
                        }
                        break;
                    case Operation.TWO:
                        switch (effect)
                        {
                            case Effect.REMOVED:
                                return Color.Red;
                            case Effect.NEUTRAL:
                                return Color.Blue;
                            case Effect.ADDED:
                                return Color.Green;
                        }
                        break;
                    case Operation.ADD:
                        switch (effect)
                        {
                            case Effect.REMOVED:
                                return Color.Red;
                            case Effect.NEUTRAL:
                                return Color.Blue;
                            case Effect.ADDED:
                                return Color.Green;
                        }
                        break;
                    case Operation.DEFAULT:
                        return Color.Black;
                }
                return Color.HotPink; // NOTE: This should never happen
            } // Determies what color the point should be
            Vector vec; // The vector which was modified
            Operation operation; // The operation used to modify it
            Effect effect; // Shows how the operation affected the given point
            public VectorEvent(Vector v, Operation op, Effect e)
            {
                vec = v;
                operation = op;
                effect = e;
            }
            public int X{ get { return vec.x; } }
            public int Y { get { return vec.y; } }
            public Vector Vec{ get { return vec; } }
            public Operation Oper { get { return operation; } }
            public Effect Eff { get { return effect; } }
        }

        public class VectorMultiset
        {
            Dictionary<Vector, int> dict;
            public VectorMultiset(){ dict = new Dictionary<Vector, int>(); }
            public void Add(Vector element)
            {
                if (dict.ContainsKey(element)) dict[element]++;
                else dict.Add(element, 1);

                LatestChanges.Push(element);
            }
            public void Remove(Vector element) // function for removing a point
            {
                try
                {
                    if (dict[element] == 1) dict.Remove(element);
                    else dict[element]--;

                    LatestChanges.Push(element);
                }
                catch (KeyNotFoundException)
                {
                    ReportError(string.Format("Key not found in dictionary: ({0}, {1})", element.x, element.y));
                }
            }
            public Dictionary<Vector, int>.Enumerator GetEnumerator() { return dict.GetEnumerator(); } // Enables looping with foreach
            public int this[Vector i]
            {
                get { return dict[i]; }
            } // Enables dictionary access
            public bool ContainsKey(Vector key){ return (dict.ContainsKey(key)); } // Wrapper for Dictionary.ContainsKey()
        } // Simulates a Vector multiset using a Dictinary<Vector, int>
        public class Result
        {
            public const int MAX_REMOVED = 20;
            public int x = 0; // Removed points
            public int cost = 0;
            public Lines lines; // The bounding lines
            public Result() { }

            public int CalculateResult()
            {
                if (Grid.Finished)
                {
                    double res = 25 * (lines.dist + (1.0 / 100) * Math.Pow(1.355, x) * (lines.dist + 1)) + cost * ((lines.dist + 1) / 50);
                    return (int)(res);
                }
                else
                {
                    return -1;
                }
            }
        } // Result class used for storing and calculating information about final result
        public class GridInfo
        {
            public enum MoveState {SHOW_REMOVED_NEUTRAL = 0, SHOW_NEUTRAL_ADDED = 1};
            public VectorMultiset points; // Contains all points currently on grid
            public Stack<Vector> pointsStack; // Contains all leftover points in the stack
            public double angle; // The final angle
            public Result res; // Contains information used for calculating final result
            bool init; // Indicates whether the grid has been initialized, i.e. if it has any data
            int nextMove = 1; // Indicates the line in OutputFile.Data, where the grid should look for information about the next move
            MoveState move_state = MoveState.SHOW_NEUTRAL_ADDED; // Indicates which points should be visualized
            bool errorState = false; // True if one of the transformations returned false
            public List<VectorEvent>[] pointsAffected; // PointsAdded[i] contains the points affected on turn i

            enum Cost{ HOMOTETHY_PLUS2 = 1, HOMOTETHY_MINUS2 = 6, TWO = 3, REPLACE = 15, DEFAULT = 0 };
            enum Operation{ HOMOTETHY_PLUS2, HOMOTETHY_MINUS2, TWO, REPLACE, DEFAULT };

            int totalCostHomotethy = 0;
            int totalCostTwo = 0;
            int totalCostReplace = 0;
            int costLastOperation = 0;

            public GridInfo()
            {
                points = new VectorMultiset();
                pointsStack = new Stack<Vector>();
                res = new Result();
                angle = -1; // unassigned
                init = false;
            } // Default constructor

            // Used for simulation
            public bool Homothety(int coefficient, Vector A, Vector B)
            {
                if (coefficient != 2 && coefficient != -2)
                {
                    errorState = true;
                    return false;
                }
                if (A == B || !points.ContainsKey(A) || !points.ContainsKey(B))
                {
                    errorState = true;
                    return false;
                }
                points.Remove(B);
                pointsAffected[nextMove - 1].Add(new VectorEvent(B, VectorEvent.Operation.HOMOTETHY, VectorEvent.Effect.REMOVED));

                B = B - A;
                B = B * coefficient;
                B = B + A;

                if (!IsPointInGrid(B.x, B.y))
                {
                    res.x++;
                    if (res.x > Result.MAX_REMOVED)
                    {
                        errorState = true;
                        return false;
                    }
                }
                else
                {
                    points.Add(B);
                    pointsAffected[nextMove - 1].Add(new VectorEvent(B, VectorEvent.Operation.HOMOTETHY, VectorEvent.Effect.ADDED));
                }
                LatestChanges.Push(A);
                pointsAffected[nextMove - 1].Add(new VectorEvent(A, VectorEvent.Operation.HOMOTETHY, VectorEvent.Effect.NEUTRAL));

                return true;
            } // realize homothety, return false if the operation is incorrect
            public bool Two(Vector A, Vector B)
            {
                if (A == B || !points.ContainsKey(A) || !points.ContainsKey(B))
                {
                    errorState = true;
                    return false;
                }
                points.Remove(A);
                pointsAffected[nextMove-1].Add(new VectorEvent(A, VectorEvent.Operation.TWO, VectorEvent.Effect.REMOVED));

                Vector C = ((A - B).Rotate(0, -1, 1, 0) * 2) + B;
                Vector D = ((A - B).Rotate(0, 1, -1, 0) * 2) + B;

                if (!IsPointInGrid(C.x, C.y))
                {
                    res.x++;
                    if (res.x > Result.MAX_REMOVED)
                    {
                        errorState = true;
                        return false;
                    }
                }
                else
                {
                    points.Add(C);
                    pointsAffected[nextMove - 1].Add(new VectorEvent(C, VectorEvent.Operation.TWO, VectorEvent.Effect.ADDED));
                }

                if (!IsPointInGrid(D.x, D.y))
                {
                    res.x++;
                    if (res.x > Result.MAX_REMOVED)
                    {
                        errorState = true;
                        return false;
                    }
                }
                else
                {
                    points.Add(D);
                    pointsAffected[nextMove - 1].Add(new VectorEvent(D, VectorEvent.Operation.TWO, VectorEvent.Effect.ADDED));
                }
                LatestChanges.Push(B);
                pointsAffected[nextMove - 1].Add(new VectorEvent(B, VectorEvent.Operation.TWO, VectorEvent.Effect.NEUTRAL));
                return true;
            } // realize operation two, return false if the operation is incorrect
            public  bool Replace(Vector A)
            {
                if (!points.ContainsKey(A))
                {
                    errorState = true;
                    return false;
                }
                if (pointsStack.Count == 0)
                {
                    errorState = true;
                    return false;
                }

                points.Remove(A);
                pointsAffected[nextMove - 1].Add(new VectorEvent(A, VectorEvent.Operation.ADD, VectorEvent.Effect.REMOVED));
                Vector B = pointsStack.Pop();
                points.Add(B);
                pointsAffected[nextMove - 1].Add(new VectorEvent(B, VectorEvent.Operation.ADD, VectorEvent.Effect.ADDED));

                return true;
            } // realize replacement, return false if the operation is invalid

            public void GetInput(string[] input, double angle)
            {
                pointsAffected = new List<VectorEvent>[Convert.ToInt32(OutputFile.Data[0]) + 2];
                for (int i = 0; i < pointsAffected.Length; ++i)
                {
                    pointsAffected[i] = new List<VectorEvent>() { };
                }
                int n = Convert.ToInt32(input[0]);
                for(int i = 1; i <= n; ++i)
                {
                    int[] coordinates = coordinates = Array.ConvertAll(input[i].Split(' '), Int32.Parse);
                    points.Add(new Vector(coordinates[0], coordinates[1]));
                    pointsAffected[0].Add(new VectorEvent(new Vector(coordinates[0], coordinates[1]), VectorEvent.Operation.DEFAULT, VectorEvent.Effect.NEUTRAL));
                }
                int m = Convert.ToInt32(input[n + 1]);
                for(int i = n + 2; i < n + 2 + m; ++i)
                {
                    int[] coordinates = coordinates = Array.ConvertAll(input[i].Split(' '), Int32.Parse);
                    pointsStack.Push(new Vector(coordinates[0], coordinates[1]));
                }
                this.angle = angle;
                init = true;
            } // Passes input to points, pointsStack
            public bool NextMove()
            {
                if (nextMove < OutputFile.Size - 1 || move_state == MoveState.SHOW_REMOVED_NEUTRAL)
                {
                    if(move_state == MoveState.SHOW_NEUTRAL_ADDED) // If finished visualizing last move
                    {
                        move_state = MoveState.SHOW_REMOVED_NEUTRAL;
                        if(pointsAffected[nextMove].Count == 0)
                        {
                            int[] op = Array.ConvertAll(OutputFile.Data[nextMove].Split(' '), Int32.Parse);
                            switch (op[0])
                            {
                                case 1:
                                    nextMove++;
                                    if(op[1] == 2)
                                    {
                                        totalCostHomotethy += (int)Cost.HOMOTETHY_PLUS2;
                                        costLastOperation = (int)Cost.HOMOTETHY_PLUS2;
                                    }
                                    else
                                    {
                                        totalCostHomotethy += (int)Cost.HOMOTETHY_MINUS2;
                                        costLastOperation = (int)Cost.HOMOTETHY_MINUS2;
                                    }
                                    return Homothety(op[1], new Vector(op[2], op[3]), new Vector(op[4], op[5]));
                                case 2:
                                    nextMove++;
                                    totalCostTwo += (int)Cost.TWO;
                                    costLastOperation = (int)Cost.TWO;
                                    return Two(new Vector(op[1], op[2]), new Vector(op[3], op[4]));
                                case 3:
                                    nextMove++;
                                    totalCostReplace += (int)Cost.REPLACE;
                                    costLastOperation = (int)Cost.REPLACE;
                                    return Replace(new Vector(op[1], op[2]));
                                default:
                                    return false;
                            }
                        }
                        else
                        {
                            Operation op = GetOpByMoveNum(nextMove);
                            switch(op)
                            {
                                case Operation.HOMOTETHY_PLUS2:
                                    totalCostHomotethy += (int)Cost.HOMOTETHY_PLUS2;
                                    costLastOperation = (int)Cost.HOMOTETHY_PLUS2;
                                    break;
                                case Operation.HOMOTETHY_MINUS2:
                                    totalCostHomotethy += (int)Cost.HOMOTETHY_MINUS2;
                                    costLastOperation = (int)Cost.HOMOTETHY_MINUS2;
                                    break;
                                case Operation.TWO:
                                    totalCostTwo += (int)Cost.TWO;
                                    costLastOperation = (int)Cost.TWO;
                                    break;
                                case Operation.REPLACE:
                                    totalCostReplace += (int)Cost.REPLACE;
                                    costLastOperation = (int)Cost.REPLACE;
                                    break;
                            }
                            foreach (var v in pointsAffected[nextMove])
                            {
                                if (v.Eff == VectorEvent.Effect.ADDED)
                                {
                                    points.Add(v.Vec);
                                }
                                else if (v.Eff == VectorEvent.Effect.REMOVED)
                                {
                                    points.Remove(v.Vec);
                                }
                                else
                                {
                                    LatestChanges.Push(v.Vec);
                                }
                            }
                            nextMove++;
                            return true;
                        }
                    }
                    else
                    {
                        move_state = MoveState.SHOW_NEUTRAL_ADDED;
                        return true;
                    }
                }
                else if(nextMove == OutputFile.Size - 1)
                {
                    res.lines = DetermineLines();
                    nextMove++;
                    return true;
                }
                else
                { 
                    return true;
                }
                
            } // Processes the next operation
            public void PrevMove()
            {
                if(nextMove > 1)
                {
                    nextMove--;
                    move_state = MoveState.SHOW_REMOVED_NEUTRAL;
                    foreach (var v in pointsAffected[nextMove])
                    {
                        if(v.Eff == VectorEvent.Effect.ADDED)
                        {
                            points.Remove(v.Vec);
                        }
                        else if(v.Eff == VectorEvent.Effect.REMOVED)
                        {
                            points.Add(v.Vec);
                        }
                        else
                        {
                            LatestChanges.Push(v.Vec);
                        }
                    }
                    Operation op = GetOpByMoveNum(nextMove);
                    Operation prevOp = GetOpByMoveNum(nextMove - 1);
                    //Debug(op.ToString());
                    switch (op)
                    {
                        case Operation.HOMOTETHY_PLUS2:
                            totalCostHomotethy -= (int)Cost.HOMOTETHY_PLUS2;
                            break;
                        case Operation.HOMOTETHY_MINUS2:
                            totalCostHomotethy -= (int)Cost.HOMOTETHY_MINUS2;
                            break;
                        case Operation.TWO:
                            totalCostTwo -= (int)Cost.TWO;
                            break;
                        case Operation.REPLACE:
                            totalCostReplace -= (int)Cost.REPLACE;
                            break;
                    }
                    switch(prevOp)
                    {
                        case Operation.HOMOTETHY_PLUS2:
                            costLastOperation = (int)Cost.HOMOTETHY_PLUS2;
                            break;
                        case Operation.HOMOTETHY_MINUS2:
                            costLastOperation = (int)Cost.HOMOTETHY_MINUS2;
                            break;
                        case Operation.TWO:
                            costLastOperation = (int)Cost.TWO;
                            break;
                        case Operation.REPLACE:
                            costLastOperation = (int)Cost.REPLACE;
                            break;
                        case Operation.DEFAULT:
                            costLastOperation = 0;
                            break;

                    }
                }
            }

            Operation GetOpByMoveNum(int move_num)
            {
                int[] temp_arr = new int[0];
                if (move_num >= 1 && move_num < OutputFile.Size - 1)
                {
                    temp_arr = Array.ConvertAll(OutputFile.Data[move_num].Split(' '), Int32.Parse);
                }
                if (temp_arr.Length == 6)
                {
                    if (temp_arr[1] == 2)
                    {
                        return Operation.HOMOTETHY_PLUS2;
                    }
                    else
                    {
                        return Operation.HOMOTETHY_MINUS2;
                    }
                }
                else if (temp_arr.Length == 5)
                {
                    return Operation.TWO;
                }
                else if (temp_arr.Length == 3)
                {
                    return Operation.REPLACE;
                }
                else
                {
                    return Operation.DEFAULT;
                }
            }

            // Used at end of execution
            Vector findEnd(DoubleVector A, double X, double Y)
            {
                DoubleVector B = new DoubleVector();
                Vector C;

                B.x = X;
                B.y = A.y - (A.x-B.x) * Math.Tan(angle);
                C = new Vector((int)Math.Round(B.x), (int)Math.Round(B.y));
                if (IsPointInGrid(C.x, C.y)) return C;

                B.y = Y;
                B.x = A.x - (A.y - B.y) / Math.Tan(angle);
                C = new Vector((int)Math.Round(B.x), (int)Math.Round(B.y));
                return C;
            } // Used to find the ends of the lines

            
            public Lines DetermineLines()
            {
                DoubleVector current, P1 = new DoubleVector(), P2 = new DoubleVector();
                double minY = 0, maxY = 0;
                bool initialized = false;

                foreach(KeyValuePair<Vector, int> entry in points)
                {
                    current = new DoubleVector((double)entry.Key.x, (double)entry.Key.y);
                    current = current.Rotate(Math.Cos(-angle), -Math.Sin(-angle), Math.Sin(-angle), Math.Cos(-angle));

                    if (!initialized)
                    {
                        minY = current.y;
                        maxY = current.y;
                        P1.x = P2.x = entry.Key.x;
                        P1.y = P2.y = entry.Key.y;
                        initialized = true;
                    }
                    else
                    {
                        if (current.y < minY)
                        {
                            minY = current.y;
                            P1.x = entry.Key.x;
                            P1.y = entry.Key.y;
                        }
                        if (current.y > maxY)
                        {
                            maxY = current.y;
                            P2.x = entry.Key.x;
                            P2.y = entry.Key.y;
                        }
                    }
                }

                Lines lines = new Lines();
                lines.dist = maxY - minY;
                if (angle < Math.PI / 2)
                {
                    lines.A = findEnd(P1, 0, 0);
                    lines.B = findEnd(P1, 1000, 700);
                    lines.C = findEnd(P2, 0, 0);
                    lines.D = findEnd(P2, 1000, 700);
                }
                else
                {
                    lines.A = findEnd(P1, 0, 700);
                    lines.B = findEnd(P1, 1000, 0);
                    lines.C = findEnd(P2, 0, 700);
                    lines.D = findEnd(P2, 1000, 0);
                }

                return lines;

            } // Returns bounding lines
            public int FinalResult{ get { res.cost = TotalCostAll; return res.CalculateResult(); } } // Calculates final result
            public bool IsInit 
            {
                get { return init; }
                set { init = value; }
            } // Getter/Setter for init
            public bool ErrorState
            {
                get { return errorState; }
            } // Getter for errorState
            public bool Finished { get { return nextMove == OutputFile.Size;  } } // Indicates if simulation has reached the end
            public int NextMoveNum { get { return nextMove; } } // Getter for nextMove
            public MoveState CurrentMoveState{ get { return move_state; } }
            public int TotalCostHomotethy { get { return totalCostHomotethy; } }
            public int TotalCostTwo { get { return totalCostTwo; } }
            public int TotalCostReplace { get { return totalCostReplace; } }
            public int TotalCostAll { get { return totalCostHomotethy + totalCostTwo + totalCostReplace; } }
            public int CostLastOperation { get { return costLastOperation; } }
        } // Grid class used for storing and processing grid information

        // Classes used for file processing - opening, reading, and validation
        public enum FPResult { FILE_VALID, FILE_INVALID, SELECT_CANCELLED }; // Used as return type when opening a file using a FileProcessor
        public abstract class FileProcessor
        {
            protected string filePath = ""; // File stream. Set when choosing file with chooseFile()
            protected string[] fileData = null; // The parsed data of the file
            protected string fileStatus = FILE_DEFAULT; // Current status of the file
            public const string FILE_INVALID = "File format is invalid";
            public const string FILE_VALID = "File is OK";
            public const string FILE_DEFAULT = "No file selected";

            // Attempts to open a file
            public FPResult OpenFile()
            {
                if (ChooseFile())
                {
                    if (ParseFile())
                    {
                        fileStatus = FILE_VALID;
                        return FPResult.FILE_VALID;
                    }
                    else
                    {
                        fileStatus = FILE_INVALID;
                        return FPResult.FILE_INVALID;
                    }
                }
                else
                {
                    return FPResult.SELECT_CANCELLED;
                }

            }
            protected abstract bool ChooseFile(); // Choose file with dialog box. Sets fileStream to target selected file.
            protected abstract bool CheckValid(string[] tempData); // Check if selected file is in valid format.
            protected abstract bool ParseFile(); // Fills fileData with processed input if input is valid
            public void WriteFile()
            {
                if (fileData != null)
                {
                    // Set a variable to the My Documents path.
                    string mydocpath = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);

                    // Write the string array to a new file named "WriteLines.txt".
                    using (StreamWriter outputFile = new StreamWriter(mydocpath + @"\WriteLines.txt"))
                    {
                        foreach (string line in fileData)
                            outputFile.WriteLine(line);
                    }
                }
            } // This function is for testing only
            public string Status
            {
                get { return fileStatus; }
            } // Getter method for fileStatus property
            public bool IsOK()
            {
                if(fileStatus == FILE_VALID)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            } // Is file format valid (return fales if no file is selected)
            public string[] Data
            {
                get { return fileData; }
            }
            public int Size { get { return fileData.Length; } }
        }
        public class InputProcessor : FileProcessor
        {
            protected override bool ChooseFile()
            {
                OpenFileDialog fileDialogBox = new OpenFileDialog()
                {
                    Title = "Choose input file",
                    InitialDirectory = Directory.GetCurrentDirectory(),
                    //InitialDirectory = Environment.SpecialFolder.MyDocuments.ToString(),
                    Filter = "In files (*.in)|*.in|All files (*.*)|*.*",
                    FilterIndex = 0,
                    RestoreDirectory = true
                };
                if (fileDialogBox.ShowDialog() == DialogResult.OK)
                {
                    try
                    {
                        if ((filePath = fileDialogBox.FileName) != "") // If a file was selected
                        {
                            return true;
                        }
                        else
                        {
                            return false;
                        }
                    }
                    catch (Exception ex)
                    {
                        ReportError("Error: Could not read file from disk. Original error: " + ex.Message);
                    }
                }
                return false;
            }
            protected override bool ParseFile()
            {
                if (filePath != "")
                {
                    string[] tempData = File.ReadAllLines(filePath);
                    if (CheckValid(tempData))
                    {
                        fileData = tempData;
                        return true;
                    }
                    else
                    {
                        return false;
                    }
                }
                else
                {
                    return false;
                }
            }
            protected override bool CheckValid(string[] tempData)
            {
                if (tempData.Length == 0)
                {
                    ReportError("Input file invalid: empty file");
                    return false;
                }
                // Read number of points
                int N;
                try
                {
                    N = Convert.ToInt32(tempData[0]);
                }
                catch (OverflowException)
                {
                    ReportError("Input file invalid: integer N too big (overflow)");
                    return false;
                }
                catch (FormatException)
                {
                    ReportError("Input file invalid: first line is not an integer");
                    return false;
                }

                // Read actual points
                for (int i = 1; i < N + 1; ++i)
                {
                    int[] coordinates;
                    try
                    {
                        coordinates = Array.ConvertAll(tempData[i].Split(' '), Int32.Parse);
                    }
                    catch (FormatException)
                    {
                        ReportError(string.Format("Input file invalid: line {0} - format invalid", i + 1));
                        return false;
                    }
                    catch (OverflowException)
                    {
                        ReportError(string.Format("Input file invalid: line {0} - integer too big (overflow)", i + 1));
                        return false;
                    }
                    if (coordinates.Length != 2)
                    {
                        if (coordinates.Length > 2)
                        {
                            ReportError(string.Format("Input file invalid: line {0} - too many arguments", i + 1));
                            return false;
                        }
                        else
                        {
                            ReportError(string.Format("Input file invalid: line {0} - not enough arguments", i + 1));
                            return false;
                        }
                    }
                    else
                    {
                        if (!IsPointInGrid(coordinates[0], coordinates[1]))
                        {
                            ReportError(string.Format("Input file invalid: line {0} - point not inside grid", i + 1));
                            return false;
                        }
                    }
                }

                // Read number of points on stack
                int M;
                try
                {
                    M = Convert.ToInt32(tempData[N + 1]);
                }
                catch (OverflowException)
                {
                    ReportError("Input file invalid: integer M too big (overflow)");
                    return false;
                }
                catch (FormatException)
                {
                    ReportError(string.Format("Input file invalid: line {0} is not an integer", N + 1));
                    return false;
                }
                // Read points on stack
                for (int i = N + 2; i < N + 2 + M; ++i)
                {
                    int[] coordinates;
                    try
                    {
                        coordinates = Array.ConvertAll(tempData[i].Split(' '), Int32.Parse);
                    }
                    catch (FormatException)
                    {
                        ReportError(string.Format("Input file invalid: line {0} - format invalid", i + 1));
                        return false;
                    }
                    catch (OverflowException)
                    {
                        ReportError(string.Format("Input file invalid: line {0} - integer too big (overflow)", i + 1));
                        return false;
                    }
                    if (coordinates.Length != 2)
                    {
                        if (coordinates.Length > 2)
                        {
                            ReportError(string.Format("Input file invalid: line {0} - too many arguments", i + 1));
                            return false;
                        }
                        else
                        {
                            ReportError(string.Format("Input file invalid: line {0} - not enough arguments", i + 1));
                            return false;
                        }
                    }
                    else
                    {
                        if (!IsPointInGrid(coordinates[0], coordinates[1]))
                        {
                            ReportError(string.Format("Input file invalid: line {0} - point not inside grid", i + 1));
                            return false;
                        }
                    }
                }

                return true;
            }
            
        }
        public class OutputProcessor : FileProcessor
        {
            double angle; // The angle, as extracted from CheckValid()
            public double Angle { get { return angle;  } }
            protected override bool ChooseFile()
            {
                OpenFileDialog fileDialogBox = new OpenFileDialog()
                {
                    Title = "Choose output file",
                    InitialDirectory = Directory.GetCurrentDirectory(),
                    //InitialDirectory = Environment.SpecialFolder.MyDocuments.ToString(),
                    Filter = "In files (*.out)|*.out|All files (*.*)|*.*",
                    FilterIndex = 0,
                    RestoreDirectory = true
                };
                if (fileDialogBox.ShowDialog() == DialogResult.OK)
                {
                    try
                    {
                        if ((filePath = fileDialogBox.FileName) != "") // If a file was selected
                        {
                            return true;
                        }
                        else
                        {
                            return false;
                        }
                    }
                    catch (Exception ex)
                    {
                        ReportError("Error: Could not read file from disk. Original error: " + ex.Message);
                    }
                }
                return false;
            }
            protected override bool ParseFile()
            {
                if (filePath != "")
                {
                    string[] tempData = File.ReadAllLines(filePath);
                    if (CheckValid(tempData))
                    {
                        fileData = tempData;
                        return true;
                    }
                    else
                    {
                        return false;
                    }
                }
                else
                {
                    return false;
                }
            }
            protected override bool CheckValid(string[] tempData)
            {
                if(tempData.Length == 0)
                {
                    ReportError("Output file invalid: empty file");
                    return false;
                }
                // Read number of operations
                int N;
                try
                {
                    N = Convert.ToInt32(tempData[0]);
                }
                catch (OverflowException)
                {
                    ReportError("Output file invalid: too many operations (overflow)");
                    return false;
                }
                catch (FormatException)
                {
                    ReportError("Output file invalid: first line is not an integer ");
                    return false;
                }
                if (N > 10000)
                {
                    ReportError("Output file invalid: too many operations (>10000)");
                    return false;
                }
                if (tempData.Length != N + 2)
                {
                    ReportError("Output file invalid: incorrect amount of lines");
                    return false;
                }

                // Read operations
                for (int i = 1; i <= N; ++i)
                {
                    int[] operation;
                    try
                    {
                        operation = Array.ConvertAll(tempData[i].Split(' '), Int32.Parse);
                    }
                    catch (FormatException)
                    {
                        ReportError(string.Format("Output file invalid: line {0} - format invalid", i + 1));
                        return false;
                    }
                    catch (OverflowException)
                    {
                        ReportError(string.Format("Output file invalid: line {0} - integer too big (overflow)", i + 1));
                        return false;
                    }
                    switch(operation[0])
                    {
                        case 1:
                            if(operation.Length != 6 || Math.Abs(operation[1]) != 2 || !IsPointInGrid(operation[2], operation[3]) || !IsPointInGrid(operation[4], operation[5]))
                            {
                                ReportError(string.Format("Output file invalid: line {0} - incorrect operation parameters", i + 1));
                                return false;
                            }
                            break;
                        case 2:
                            if(operation.Length != 5 || !IsPointInGrid(operation[1], operation[2]) || !IsPointInGrid(operation[3], operation[4]))
                            {
                                ReportError(string.Format("Output file invalid: line {0} - incorrect operation parameters", i + 1));
                                return false;
                            }
                            break;
                        case 3:
                            if (operation.Length != 3 || !IsPointInGrid(operation[1], operation[2]))
                            {
                                ReportError(string.Format("Output file invalid: line {0} - incorrect operation parameters", i + 1));
                                return false;
                            }
                            break;
                        default:
                            ReportError(string.Format("Output file invalid: line {0} - no such operation", i + 1));
                            return false;
                    }
                }

                // Check final angle for correct number of significant digits
                double angle;
                try
                {
                    angle = Convert.ToDouble(tempData[N+1], CultureInfo.InvariantCulture);
                }
                catch (OverflowException)
                {
                    ReportError("Output file invalid: last line - angle too big (overflow)");
                    return false;
                }
                catch (FormatException)
                {
                    ReportError("Output file invalid: last line - angle format incorrect");
                    return false;
                }
                if(angle < 0 || angle > Math.PI)
                {
                    ReportError("Angle not in range [0, PI]");
                    return false;
                }
             
                // MANUAL TRUNCATE
                decimal truncated_angle = (decimal)angle;
                truncated_angle *= (decimal)Math.Pow(10, 7);
                truncated_angle = (int)truncated_angle;
                truncated_angle /= (decimal)Math.Pow(10, 7);
                // END MANUAL TRUNCATE

                if ( (decimal)angle != truncated_angle )
                {
                    ReportError("Output file invalid: last line - angle has more than 8 digits");
                    return false;
                }
                else
                {
                    this.angle = angle;
                }

                return true;
            } // Checks if file is in valid format. Does NOT check if file produces WRONG ANSWER.
        }

        // Process file I/O
        private void inputButton_Click(object sender, EventArgs e)
        {
            pauseButton.PerformClick(); // Pause execution if running
            if(Grid.IsInit == true) // If already intialized before
            {
                switch(InputFile.OpenFile())
                {
                    case FPResult.FILE_VALID:
                    case FPResult.FILE_INVALID:
                        Grid = new GridInfo();
                        panel1.Invalidate();
                        if (InputFile.IsOK() && OutputFile.IsOK())
                        {
                            Grid.GetInput(InputFile.Data, OutputFile.Angle);
                        }
                        break;
                    case FPResult.SELECT_CANCELLED:
                        break;
                }
            }
            else
            {
                InputFile.OpenFile();
                if(InputFile.IsOK() && OutputFile.IsOK())
                {
                    Grid.GetInput(InputFile.Data, OutputFile.Angle);
                }
            }
        } // Select input (.in) file
        private void outputButton_Click(object sender, EventArgs e)
        {
            pauseButton.PerformClick();
            if( Grid.IsInit )
            {
                switch (OutputFile.OpenFile())
                {
                    case FPResult.FILE_VALID:
                    case FPResult.FILE_INVALID:
                        Grid = new GridInfo();
                        if (InputFile.IsOK() && OutputFile.IsOK())
                        {
                            Grid.GetInput(InputFile.Data, OutputFile.Angle);
                        }
                        panel1.Invalidate();
                        break;
                    case FPResult.SELECT_CANCELLED:
                        break;
                }
            }
            else
            {
                OutputFile.OpenFile();
                if (InputFile.IsOK() && OutputFile.IsOK())
                {
                    Grid.GetInput(InputFile.Data, OutputFile.Angle);
                }
            }
            
        } // Select output (.out) file

        // Drawing grid
        public const int GRID_SPACING = 100; // Space between lines in pixels
        public const int GRID_WIDTH = 1000; // Width of grid
        public const int GRID_HEIGHT = 700; // Height of grid
        public const int POINT_WIDTH = 5;
        static bool IsPointInGrid(int x, int y)
        {
            if (x < 0 || x > GRID_WIDTH || y < 0 || y > GRID_HEIGHT)
            {
                return false;
            }
            else
            {
                return true;
            }
        } // Check if point is in grid (grid dimensions are GRID_WIDTH and GRID_HEIGTH constants)

        static int inverseOrdinate(int y)
        {
            return GRID_HEIGHT - y;
        }

        static int centerCoordinate(int coordinate)
        {
            return coordinate - POINT_WIDTH / 2;
        }

        private void Invalidate_Stack()
        {
            while (LatestChanges.Count != 0)
            {
                var v = LatestChanges.Pop();
                Rectangle point = new Rectangle(centerCoordinate(v.x), centerCoordinate(inverseOrdinate(v.y)), POINT_WIDTH + 1, POINT_WIDTH + 1); // previously rectangle was POINT_WIDTH on POINT_WIDTH which is not enough
                panel1.Invalidate(point);
            }
            if (!Grid.Finished) // This is so that reloading a file doesn't cause a bug
            {
                FinishedVisualization = false;
            }
            if (Grid.Finished && !FinishedVisualization)
            {
                panel1.Invalidate();
                FinishedVisualization = true;
            }
        } // Invalidates all points in LatestChanges (and invalidates panel if visualization finished, so lines are drawn)
        private void Visualizer_Paint(object sender, PaintEventArgs e)
        {
            if (InputFile.IsOK() && OutputFile.IsOK())
            {
                nextButton.Visible = true;
                prevButton.Visible = true;
                playButton.Visible = true;
                pauseButton.Visible = true;

                UIText1.Visible = true;
                UIText1_1.Visible = true;
                nextMoveInterval.Visible = true;
                UIText2.Visible = true;
                lastMoveCost.Visible = true;
                UIText3.Visible = true;
                op1CostTotal.Visible = true;
                UIText4.Visible = true;
                op2CostTotal.Visible = true;
                UIText5.Visible = true;
                op3CostTotal.Visible = true;
                UIText6.Visible = true;
                costTotal.Visible = true;
                UIText7.Visible = true;
                removedPointsCount.Visible = true;
                UIText8.Visible = true;


                op1CostTotal.Text = Grid.TotalCostHomotethy.ToString();
                op2CostTotal.Text = Grid.TotalCostTwo.ToString();
                op3CostTotal.Text = Grid.TotalCostReplace.ToString();
                costTotal.Text = Grid.TotalCostAll.ToString();
                removedPointsCount.Text = Grid.res.x.ToString();
                lastMoveCost.Text = Grid.CostLastOperation.ToString();

                if (Grid.NextMoveNum > 1) // If a move has been performed, show it
                {
                    lastMoveSummary.Visible = true;
                    lastMoveSummary.Text = OutputFile.Data[Grid.NextMoveNum - 1];
                }
                else
                {
                    lastMoveSummary.Visible = false;
                }
                while(LatestChanges.Count != 0)
                {
                    var v = LatestChanges.Pop();
                    Rectangle point = new Rectangle(centerCoordinate(v.x), centerCoordinate(inverseOrdinate(v.y)), POINT_WIDTH + 1, POINT_WIDTH + 1); // previously rectangle was 5x5 which is not enough
                    panel1.Invalidate(point);
                }
                if(!Grid.Finished) // This is so that reloading a file doesn't cause a bug
                {
                    FinishedVisualization = false;
                    UIText9.Visible = false;
                    finalScoreLabel.Visible = false;
                    UIText10.Visible = false;
                    finalDistanceLabel.Visible = false;
                }
                if(Grid.Finished)
                {
                    UIText9.Visible = true;
                    finalScoreLabel.Visible = true;
                    finalScoreLabel.Text = Grid.FinalResult.ToString();
                    UIText10.Visible = true;
                    finalDistanceLabel.Visible = true;
                    finalDistanceLabel.Text = ((int)Grid.res.lines.dist).ToString();
                }
                if(Grid.Finished && !FinishedVisualization)
                {
                    panel1.Invalidate();
                    FinishedVisualization = true;
                }
            }
            else
            {
                nextButton.Visible = false;
                prevButton.Visible = false;
                playButton.Visible = false;
                pauseButton.Visible = false;

                UIText1.Visible = false;
                UIText1_1.Visible = false;
                nextMoveInterval.Visible = false;
                UIText2.Visible = false;
                lastMoveCost.Visible = false;
                UIText3.Visible = false;
                op1CostTotal.Visible = false;
                UIText4.Visible = false;
                op2CostTotal.Visible = false;
                UIText5.Visible = false;
                op3CostTotal.Visible = false;
                UIText6.Visible = false;
                costTotal.Visible = false;
                UIText7.Visible = false;
                lastMoveSummary.Visible = false;
                removedPointsCount.Visible = false;
                UIText8.Visible = false;
            }
        } // Controls which parts of the UI are visible

        private void panel1_Paint(object sender, PaintEventArgs e)
        {
            e.Graphics.SmoothingMode = SmoothingMode.AntiAlias;
            Pen backgroundPen = new Pen(Color.FromArgb(40, 0, 0, 0), 1);
            
            // Draw grid
            for (int i = 0; i * GRID_SPACING < GRID_HEIGHT; ++i)
            {
                e.Graphics.DrawLine(backgroundPen, 0, i * GRID_SPACING, GRID_WIDTH, i * GRID_SPACING);
            }
            for (int i = 0; i * GRID_SPACING <= GRID_WIDTH; ++i)
            {
                e.Graphics.DrawLine(backgroundPen, i * GRID_SPACING, 0, i * GRID_SPACING, GRID_HEIGHT);
            }
            // Draw points
            Brush pointBrush = new SolidBrush(Color.FromArgb(255, 0, 0, 0));
            if(Grid.IsInit)
            {
                List<Vector> affectedVectors = new List<Vector>(); // Vectors affected by last operation
                if (Grid.NextMoveNum > 1 && Grid.NextMoveNum < OutputFile.Size)
                {
                    foreach (var aff in Grid.pointsAffected[Grid.NextMoveNum - 1])
                    {
                        affectedVectors.Add(aff.Vec);
                    }
                }
                foreach (var v in Grid.points)
                {
                    if (!affectedVectors.Contains(v.Key))
                    {
                        Rectangle point = new Rectangle(centerCoordinate(v.Key.x), centerCoordinate(inverseOrdinate(v.Key.y)), POINT_WIDTH, POINT_WIDTH);
                        e.Graphics.FillRectangle(pointBrush, point);
                    }
                }
                if (Grid.NextMoveNum > 1 && Grid.NextMoveNum < OutputFile.Size)
                {
                    foreach (var v in Grid.pointsAffected[Grid.NextMoveNum - 1])
                    {
                        Rectangle point = new Rectangle(centerCoordinate(v.X), centerCoordinate(inverseOrdinate(v.Y)), POINT_WIDTH, POINT_WIDTH);
                        if (Grid.CurrentMoveState == GridInfo.MoveState.SHOW_REMOVED_NEUTRAL)
                        {
                            if (v.Eff == VectorEvent.Effect.REMOVED || v.Eff == VectorEvent.Effect.NEUTRAL)
                            {
                                e.Graphics.FillRectangle(new SolidBrush(v.GetColor()), point);
                            }
                        }
                        else
                        {
                            if (v.Eff == VectorEvent.Effect.ADDED || v.Eff == VectorEvent.Effect.NEUTRAL)
                            {
                                e.Graphics.FillRectangle(new SolidBrush(v.GetColor()), point);
                            }
                        }
                        LatestChanges.Push(v.Vec);
                    }
                }
                if (Grid.Finished)
                {
                    Pen linePen = new Pen(Color.DarkBlue, 2);
                    Grid.res.lines.A.y = inverseOrdinate(Grid.res.lines.A.y);
                    Grid.res.lines.B.y = inverseOrdinate(Grid.res.lines.B.y);
                    Grid.res.lines.C.y = inverseOrdinate(Grid.res.lines.C.y);
                    Grid.res.lines.D.y = inverseOrdinate(Grid.res.lines.D.y);
                    e.Graphics.DrawLine(linePen, Grid.res.lines.A.ToPoint(), Grid.res.lines.B.ToPoint());
                    e.Graphics.DrawLine(linePen, Grid.res.lines.C.ToPoint(), Grid.res.lines.D.ToPoint());
                }
            }
        } // Draws grid
        // Drawing interface
        private void inputButton_Paint(object sender, PaintEventArgs e)
        {
            if(InputFile.Status == FileProcessor.FILE_VALID)
            {
                inputButton.Text = "Input - OK";
            }
            else if(InputFile.Status == FileProcessor.FILE_INVALID)
            {
                inputButton.Text = "Input - BAD";
            }
            else
            {
                inputButton.Text = "Input";
            }
        } // Selects proper text for inputButton based on InputFile.Status
        private void outputButton_Paint(object sender, PaintEventArgs e)
        {
            if (OutputFile.Status == FileProcessor.FILE_VALID)
            {
                outputButton.Text = "Output - OK";
            }
            else if (OutputFile.Status == FileProcessor.FILE_INVALID)
            {
                outputButton.Text = "Output - BAD";
            }
            else
            {
                outputButton.Text = "Output";
            }
        } // Selects proper text for outputButton based on OutputFile.Status
        
        // Animation handling
        int debug_ticks = 0;
        private void nextMoveInterval_ValueChanged(object sender, EventArgs e)
        {
            if(nextMoveInterval.Value > 0)
            {
                MoveTimer.Interval = (int)nextMoveInterval.Value;
            }
        } // Sets the tick interval of MoveTimer
        private void MoveTimer_Tick(object sender, EventArgs e)
        {
            ++debug_ticks;
            nextButton.PerformClick();
            if(Grid.Finished)
            {
                MoveTimer.Stop();
            }
            DebugLabel.Text = Convert.ToString(debug_ticks);
        } // Simulates a click of nextButton
        private void pauseButton_Click(object sender, EventArgs e)
        {
            MoveTimer.Stop();
            nextMoveInterval.Minimum = 0;
        } // Pauses MoveTimer and thus pauses animated execution
        private void playButton_Click(object sender, EventArgs e)
        {
            if(nextMoveInterval.Value > 0)
            {
                MoveTimer.Start();
                nextMoveInterval.Minimum = 1; // Prevents user from setting too low animation time
            }
        } // Starts MoveTimer and thus starts animated exectuion
        private void nextButton_Click(object sender, EventArgs e)
        {
            if(Grid.IsInit && !Grid.ErrorState)
            {
                if (!Grid.NextMove())
                {
                    ReportError("Latest move was invalid!");
                }
                Invalidate(); // Forces call to Visualizer_Paint() (probably)
            }
        } // Performs next move
        private void prevButton_Click(object sender, EventArgs e)
        {
            if (Grid.IsInit && !Grid.ErrorState)
            {
                Grid.PrevMove();
                Refresh();
            }
        } // Reverses effects of last move
        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            /* Overrides arrow keys */

            if (keyData == Keys.Right)
            {
                nextButton.PerformClick();
                return true;
            }
            if (keyData == Keys.Left)
            {
                prevButton.PerformClick();
                return true;
            }
            if(keyData == Keys.Space)
            {
                if(!MoveTimer.Enabled)
                {
                    playButton.PerformClick();
                }
                else
                {
                    pauseButton.PerformClick();
                }
                return true;
            }
            return base.ProcessCmdKey(ref msg, keyData);
        } // Overrides arrow keys so that they call prevButton and nextButton and SPACEBAR to call Play/Pause

        private void testgenButton_Click(object sender, EventArgs e)
        {
            Form2 testgenDialog = new Form2();
            if(testgenDialog.ShowDialog() == DialogResult.OK)
            {
                MessageBox.Show("Test generated!", "Done", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            testgenDialog.Dispose();
        } // Opens test generator dialog
    }
}
