﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;

namespace CodeIT_S6_Visualizer
{
    public partial class Form2 : Form
    {
        public static void Debug(string message = "Test")
        {
            MessageBox.Show(message, "Debug", MessageBoxButtons.OK, MessageBoxIcon.Information);
        } // Wrapper for MessageBox.Show(). Should be used for debug purposes.
        public static void ReportError(string error_message)
        {
            MessageBox.Show(error_message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
        } // Wrapper for MessageBox.Show(). Should be used when reporting errors.

        public Form2()
        {
            InitializeComponent();
            UIText1.Visible = true;
            UIText2.Visible = true;
            UIText3.Visible = true;
            UIText4.Visible = true;
            UIText5.Visible = true;
            UIText6.Visible = true;
            UIText7.Visible = true;
            UIText8.Visible = true;
            UIText9.Visible = true;
        } // Initializer

        bool IsValidFilePath(string file_path)
        {
            FileStream fs = null;
            try
            {
                fs = File.Open(file_path, FileMode.OpenOrCreate);
            }
            catch(SystemException)
            {
                return false;
            }
            finally
            {
                if( fs != null )
                {
                    fs.Close();
                }
            }
            return true;
            
        } // Attemts to open or create file at file_path. Return false if exception occurs and true otherwise
        bool IsValidInputInteger(string s, int lower_bound, int upper_bound)
        {
            if(s == "")
            {
                return false;
            }
            else
            {
                try
                {
                    int val = Convert.ToInt32(s);
                    if (val < lower_bound || val > upper_bound)
                    {
                        return false;
                    }
                    else
                    {
                        return true;
                    }
                }
                catch(OverflowException)
                {
                    return false;
                }
            }
        } // Checks if s is not empty and is within [lower_bound,upper_bound]
        bool VerifyInput()
        {
            if (!IsValidFilePath(FilePathInputBox.Text))
            {
                ReportError("File path is invalid!");
                return false;
            }
            else if (!IsValidInputInteger(MinPointsInputBox.Text, 1, 100)
                 || !IsValidInputInteger(MaxPointsInputBox.Text, Convert.ToInt32(MinPointsInputBox.Text), 100)
                 || !IsValidInputInteger(LeftBoundInputBox.Text, 0, 1000)
                 || !IsValidInputInteger(RightBoundInputBox.Text, Convert.ToInt32(LeftBoundInputBox.Text), 1000)
                 || !IsValidInputInteger(BottomBoundInputBox.Text, 0, 700)
                 || !IsValidInputInteger(TopBoundInputBox.Text, Convert.ToInt32(BottomBoundInputBox.Text), 700)
                 || !IsValidInputInteger(SeedInputBox.Text, 0, Int32.MaxValue))
                {
                ReportError("Incorrectly given input!");
                return false;
            }
            else if(!IsValidInputInteger(PtsOnStackInputBox.Text, 0, 20))
            {
                ReportError("Too many points on stack! (Max 20)");
                return false;
            }
            else
            {
                return true;
            }
        } // Verifies if file path is correct and if point ranges are within bounds

        private void GenerateTestButton_Click(object sender, EventArgs e)
        {
            if(VerifyInput())
            {
                generateTest(FilePathInputBox.Text, Convert.ToInt32(MinPointsInputBox.Text), Convert.ToInt32(MaxPointsInputBox.Text)
                        , Convert.ToInt32(LeftBoundInputBox.Text), Convert.ToInt32(RightBoundInputBox.Text)
                        , Convert.ToInt32(BottomBoundInputBox.Text), Convert.ToInt32(TopBoundInputBox.Text), Convert.ToInt32(PtsOnStackInputBox.Text)
                        , Convert.ToInt32(SeedInputBox.Text));

                DialogResult = DialogResult.OK;
                Close();
            }
        } // Generates test and closes dialog box if input is valid. Shows error message otherwise
        private void generateTest(string file_path, int minPts, int maxPts, int leftBound, int rightBound, int botttomBound, int topBound, int ptsOnStack, int seed)
        {
            Random randGen;
            if(seed != 0)
            {
                randGen = new Random(seed);
            }
            else
            {
                randGen = new Random();
            }

            int numPoints = randGen.Next(minPts, maxPts);
            Point[] pts = new Point[numPoints];
            for(int i = 0; i < numPoints; ++i)
            {
                int x = randGen.Next(leftBound, rightBound);
                int y = randGen.Next(botttomBound, topBound);
                pts[i] = new Point(x, y);
            }
            Point[] pts_stack = new Point[ptsOnStack];
            for(int i = 0; i < ptsOnStack; ++i)
            {
                int x = randGen.Next(leftBound, rightBound);
                int y = randGen.Next(botttomBound, topBound);
                pts_stack[i] = new Point(x, y);
            }
            using (StreamWriter outputFile = new StreamWriter(file_path))
            {
                outputFile.WriteLine(numPoints);
                foreach (Point pt in pts)
                {
                    outputFile.WriteLine(string.Format("{0} {1}", pt.X, pt.Y));
                }
                outputFile.WriteLine(ptsOnStack);
                foreach(Point pt in pts_stack)
                {
                    outputFile.WriteLine(string.Format("{0} {1}", pt.X, pt.Y));
                }  
            }
        } // Generates a single testcase

        // Only allow numbers to be input into text boxes
        private void MinPointsInputBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            // If input is not a number (or a conrol key, e.g. up/down arrow, ESC, etc.), ignore it
            if (!char.IsControl(e.KeyChar) && !char.IsDigit(e.KeyChar))
            {
                e.Handled = true;
            }
        }
        private void MaxPointsInputBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            // If input is not a number (or a conrol key, e.g. up/down arrow, ESC, etc.), ignore it
            if (!char.IsControl(e.KeyChar) && !char.IsDigit(e.KeyChar))
            {
                e.Handled = true;
            }
        }
        private void LeftBoundInputBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            // If input is not a number (or a conrol key, e.g. up/down arrow, ESC, etc.), ignore it
            if (!char.IsControl(e.KeyChar) && !char.IsDigit(e.KeyChar))
            {
                e.Handled = true;
            }
        }
        private void RightBoundInputBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            // If input is not a number (or a conrol key, e.g. up/down arrow, ESC, etc.), ignore it
            if (!char.IsControl(e.KeyChar) && !char.IsDigit(e.KeyChar))
            {
                e.Handled = true;
            }
        }
        private void ButtomBoundInputBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            // If input is not a number (or a conrol key, e.g. up/down arrow, ESC, etc.), ignore it
            if (!char.IsControl(e.KeyChar) && !char.IsDigit(e.KeyChar))
            {
                e.Handled = true;
            }
        }
        private void TopBoundInputBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            // If input is not a number (or a conrol key, e.g. up/down arrow, ESC, etc.), ignore it
            if (!char.IsControl(e.KeyChar) && !char.IsDigit(e.KeyChar))
            {
                e.Handled = true;
            }
        }
        private void PtsOnStackInputBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            // If input is not a number (or a conrol key, e.g. up/down arrow, ESC, etc.), ignore it
            if (!char.IsControl(e.KeyChar) && !char.IsDigit(e.KeyChar))
            {
                e.Handled = true;
            }
        }
        private void SeedInputBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            // If input is not a number (or a conrol key, e.g. up/down arrow, ESC, etc.), ignore it
            if (!char.IsControl(e.KeyChar) && !char.IsDigit(e.KeyChar))
            {
                e.Handled = true;
            }
        }

        private void Form2_Load(object sender, EventArgs e)
        {

        }
    }
}
